@@ -17,6 +17,11 @@
 
 // A libpq-based PostgreSQL driver for ADBC.
 
+// For #warning Please include winsock2.h before windows.h on RTools/msys2
+#ifdef _WIN32
+#include <winsock2.h>
+#endif
+
 #include <cstring>
 #include <memory>
 
@@ -25,8 +30,10 @@
 #include "connection.h"
 #include "database.h"
 #include "driver/common/utils.h"
+#include "driver/framework/status.h"
 #include "statement.h"
 
+using adbc::driver::Status;
 using adbcpq::PostgresConnection;
 using adbcpq::PostgresDatabase;
 using adbcpq::PostgresStatement;
@@ -56,20 +63,44 @@ const struct AdbcError* PostgresErrorFromArrayStream(struct ArrowArrayStream* st
   // Currently only valid for TupleReader
   return adbcpq::TupleReader::ErrorFromArrayStream(stream, status);
 }
+
+int PostgresErrorGetDetailCount(const struct AdbcError* error) {
+  if (InternalAdbcIsCommonError(error)) {
+    return InternalAdbcCommonErrorGetDetailCount(error);
+  }
+
+  if (error->vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA) {
+    return 0;
+  }
+
+  auto error_obj = reinterpret_cast<Status*>(error->private_data);
+  return error_obj->CDetailCount();
+}
+
+struct AdbcErrorDetail PostgresErrorGetDetail(const struct AdbcError* error, int index) {
+  if (InternalAdbcIsCommonError(error)) {
+    return InternalAdbcCommonErrorGetDetail(error, index);
+  }
+
+  auto error_obj = reinterpret_cast<Status*>(error->private_data);
+  return error_obj->CDetail(index);
+}
 }  // namespace
 
+#if !defined(ADBC_NO_COMMON_ENTRYPOINTS)
 int AdbcErrorGetDetailCount(const struct AdbcError* error) {
-  return CommonErrorGetDetailCount(error);
+  return PostgresErrorGetDetailCount(error);
 }
 
 struct AdbcErrorDetail AdbcErrorGetDetail(const struct AdbcError* error, int index) {
-  return CommonErrorGetDetail(error, index);
+  return PostgresErrorGetDetail(error, index);
 }
 
 const struct AdbcError* AdbcErrorFromArrayStream(struct ArrowArrayStream* stream,
                                                  AdbcStatusCode* status) {
   return PostgresErrorFromArrayStream(stream, status);
 }
+#endif  // ADBC_NO_COMMON_ENTRYPOINTS
 
 // ---------------------------------------------------------------------
 // AdbcDatabase
@@ -85,11 +116,11 @@ AdbcStatusCode PostgresDatabaseInit(struct AdbcDatabase* database,
 AdbcStatusCode PostgresDatabaseNew(struct AdbcDatabase* database,
                                    struct AdbcError* error) {
   if (!database) {
-    SetError(error, "%s", "[libpq] database must not be null");
+    InternalAdbcSetError(error, "%s", "[libpq] database must not be null");
     return ADBC_STATUS_INVALID_STATE;
   }
   if (database->private_data) {
-    SetError(error, "%s", "[libpq] database is already initialized");
+    InternalAdbcSetError(error, "%s", "[libpq] database is already initialized");
     return ADBC_STATUS_INVALID_STATE;
   }
   auto impl = std::make_shared<PostgresDatabase>();
@@ -171,6 +202,7 @@ AdbcStatusCode PostgresDatabaseSetOptionInt(struct AdbcDatabase* database,
 }
 }  // namespace
 
+#if !defined(ADBC_NO_COMMON_ENTRYPOINTS)
 AdbcStatusCode AdbcDatabaseGetOption(struct AdbcDatabase* database, const char* key,
                                      char* value, size_t* length,
                                      struct AdbcError* error) {
@@ -226,6 +258,7 @@ AdbcStatusCode AdbcDatabaseSetOptionDouble(struct AdbcDatabase* database, const
                                            double value, struct AdbcError* error) {
   return PostgresDatabaseSetOptionDouble(database, key, value, error);
 }
+#endif  // ADBC_NO_COMMON_ENTRYPOINTS
 
 // ---------------------------------------------------------------------
 // AdbcConnection
@@ -426,6 +459,7 @@ AdbcStatusCode PostgresConnectionSetOptionInt(struct AdbcConnection* connection,
 
 }  // namespace
 
+#if !defined(ADBC_NO_COMMON_ENTRYPOINTS)
 AdbcStatusCode AdbcConnectionCancel(struct AdbcConnection* connection,
                                     struct AdbcError* error) {
   return PostgresConnectionCancel(connection, error);
@@ -560,6 +594,7 @@ AdbcStatusCode AdbcConnectionSetOptionDouble(struct AdbcConnection* connection,
                                              struct AdbcError* error) {
   return PostgresConnectionSetOptionDouble(connection, key, value, error);
 }
+#endif  // ADBC_NO_COMMON_ENTRYPOINTS
 
 // ---------------------------------------------------------------------
 // AdbcStatement
@@ -737,6 +772,7 @@ AdbcStatusCode PostgresStatementSetSqlQuery(struct AdbcStatement* statement,
 }
 }  // namespace
 
+#if !defined(ADBC_NO_COMMON_ENTRYPOINTS)
 AdbcStatusCode AdbcStatementBind(struct AdbcStatement* statement,
                                  struct ArrowArray* values, struct ArrowSchema* schema,
                                  struct AdbcError* error) {
@@ -846,11 +882,12 @@ AdbcStatusCode AdbcStatementSetSqlQuery(struct AdbcStatement* statement,
                                         const char* query, struct AdbcError* error) {
   return PostgresStatementSetSqlQuery(statement, query, error);
 }
+#endif  // ADBC_NO_COMMON_ENTRYPOINTS
 
 extern "C" {
 ADBC_EXPORT
-AdbcStatusCode PostgresqlDriverInit(int version, void* raw_driver,
-                                    struct AdbcError* error) {
+AdbcStatusCode AdbcDriverPostgresqlInit(int version, void* raw_driver,
+                                        struct AdbcError* error) {
   if (version != ADBC_VERSION_1_0_0 && version != ADBC_VERSION_1_1_0) {
     return ADBC_STATUS_NOT_IMPLEMENTED;
   }
@@ -860,8 +897,8 @@ AdbcStatusCode PostgresqlDriverInit(int version, void* raw_driver,
   if (version >= ADBC_VERSION_1_1_0) {
     std::memset(driver, 0, ADBC_DRIVER_1_1_0_SIZE);
 
-    driver->ErrorGetDetailCount = CommonErrorGetDetailCount;
-    driver->ErrorGetDetail = CommonErrorGetDetail;
+    driver->ErrorGetDetailCount = PostgresErrorGetDetailCount;
+    driver->ErrorGetDetail = PostgresErrorGetDetail;
     driver->ErrorFromArrayStream = PostgresErrorFromArrayStream;
 
     driver->DatabaseGetOption = PostgresDatabaseGetOption;
@@ -927,8 +964,10 @@ AdbcStatusCode PostgresqlDriverInit(int version, void* raw_driver,
   return ADBC_STATUS_OK;
 }
 
+#if !defined(ADBC_NO_COMMON_ENTRYPOINTS)
 ADBC_EXPORT
 AdbcStatusCode AdbcDriverInit(int version, void* raw_driver, struct AdbcError* error) {
-  return PostgresqlDriverInit(version, raw_driver, error);
+  return AdbcDriverPostgresqlInit(version, raw_driver, error);
 }
+#endif  // ADBC_NO_COMMON_ENTRYPOINTS
 }
